"""Rest api for lamin instance.

Import the package::

   import lndb_rest

This is the complete API reference:

.. autosummary::
   :toctree: .

   example_function
   ExampleClass
"""

__version__ = "0.0.1"  # denote a pre-release for 0.1.0 with 0.1rc1

from ._core import ExampleClass, example_function  # noqa



"""API routers."""


from .account import router  # noqa
from .introspection import router  # noqa
from .run import router  # noqa
from .instance import router  # noqa
from .file import router  # noqa



import importlib
from typing import Dict, List

import lamindb as ln
import numpy as np
from fastapi import APIRouter
from lamindb.schema._table import table_meta
from lndb import settings
from lndb import settings as _settings
from lndb.dev._setup_schema import get_schema_module_name

router = APIRouter(prefix="/introspection")

@router.get("/")
async def get_db_schema() -> Dict[str, List[str]]:
    # the following is based on `lamindb.view()`
    schema_names = ["core"] + list(_settings.instance.schema)
    all_orms = {}
    for schema_name in schema_names:
        schema_module = importlib.import_module(get_schema_module_name(schema_name))
        orms = [
            i
            for i in schema_module.__dict__.values()
            if i.__class__.__name__ == "SQLModelMetaclass" and hasattr(i, "__table__")
        ]
        # make an entry for the schema module
        all_orms[schema_name] = {orm.__name__ for orm in orms}
        # alternatively, use pydantic fields
        # for orm in orms:
        #     # capture information about all fields for every ORM
        #     all_orms[schema_name][orm.__name__] = orm.__fields__
        # alternatively, use sqlalchemy tables
        # ...
    return all_orms


@router.get("/{schema_name}/{table_name}")
async def get_table(schema_name: str, table_name: str):
    table_metaclass = get_table_metaclass(schema_name, table_name)

    # Rows
    table_df = (
        ln
        .select(table_metaclass)
        .df()
        .replace({np.nan: None})
        .reset_index()
    )
    rows = table_df.to_dict(orient="records")

    # Schema
    table_object = ln.schema._core.get_table_object(f"{schema_name}.{table_name}")
    schema = ln.schema._core.get_table_metadata_as_dict(table_object)

    return {"schema": schema, "rows": rows}


def get_table_metaclass(schema_name, table_name):
    if settings.instance.dialect == "sqlite":
        return table_meta.get_model(f"{schema_name}.{table_name}")
    else:
        return table_meta.get_model(table_name)



import os

import uvicorn
from fastapi import FastAPI
from fastapi.testclient import TestClient
from starlette.middleware.cors import CORSMiddleware

from lndb_rest.routers import account, instance, introspection, run, file

port = int(os.getenv("PORT", 8000))

app = FastAPI()

app.add_middleware(
    CORSMiddleware,
    allow_origins=["*"],
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)

app.include_router(account.router)
app.include_router(instance.router)
app.include_router(introspection.router)
app.include_router(run.router)
app.include_router(file.router)

client = TestClient(app)

if __name__ == "__main__":
    uvicorn.run("main:app", host="0.0.0.0", port=port, reload=True)



from fastapi import APIRouter
from lndb import settings
from lnhub_rest.routers.instance import (
    get_instance_by_name as get_instance_by_name_base,
)

router = APIRouter(prefix="/instance")


@router.get("/")
def get_current_instance():
    instance = get_instance_by_name_base(
        settings.instance._owner, settings.instance._name, f"Bearer {settings.user.access_token}"
    )
    return {
        **instance,
        "storage": {
            "root": "..."
        }
    }


@router.get("/settings")
def get_current_instance_settings():
    return {
        "owner": settings.instance._owner,
        "name": settings.instance._name,
        "storage": settings.instance._storage,
        "db": settings.instance._db,
        "schema_str": settings.instance._schema_str,
    }



import lamindb as ln
from fastapi import APIRouter
from fastapi import UploadFile

router = APIRouter(prefix="/file")


@router.get("/")
async def get_files(order_by: str = None, desc: bool = True):

    if order_by == "name":
        order_by_exp = ln.File.name
    else:
        order_by_exp = ln.File.created_at

    if desc is True:
        order_by_exp = order_by_exp.desc()

    files = (
        ln.select(ln.File.id, ln.File, ln.Run, ln.User, ln.Transform)
        .join(ln.Run, ln.Run.id == ln.File.source_id)
        .join(ln.User, ln.User.id == ln.Run.created_by)
        .join(ln.Transform, (ln.Transform.id == ln.Run.transform_id), (ln.Transform.v == ln.Run.transform_v))
        .order_by(order_by_exp)
    ).all()

    return files


@router.get("/{file_id}")
async def get_file(file_id: str):

    file = (
        ln.select(ln.File.id, ln.File, ln.Run, ln.User, ln.Transform)
        .join(ln.Run, ln.Run.id == ln.File.source_id)
        .join(ln.User, ln.User.id == ln.Run.created_by)
        .join(ln.Transform, (ln.Transform.id == ln.Run.transform_id), (ln.Transform.v == ln.Run.transform_v))
        .where(ln.File.id == file_id)
    ).one()

    return file


@router.post("/add")
async def add_file(uploadFile: UploadFile):
    transform = ln.add(ln.Transform(name="transform_1", v="1"))
    run = ln.Run(transform=transform)
    contents = await uploadFile.read()
    with open(uploadFile.filename, "wb") as binary_file:
        binary_file.write(contents)
    file = ln.File(uploadFile.filename, source=run)
    ln.add(file)
    return {"filename": uploadFile.filename}



def example_function(column_name: str) -> str:
    """Lower case your input string.

    Args:
        column_name: Column name to transform to lower case.

    Returns:
        The lower-cased column name.
    """
    return column_name.lower()


class ExampleClass:
    """Awesome class."""

    def __init__(self, value: int):
        print("initializing")

    def bar(self) -> str:
        """Bar function."""
        return "hello"

    @property
    def foo(self) -> str:
        """Foo property."""
        return "hello"



import argparse
import subprocess

from lndb._settings import settings
from lndb._settings_store import settings_dir

description_cli = "CLI for lndb-app."
parser = argparse.ArgumentParser(
    description=description_cli, formatter_class=argparse.RawTextHelpFormatter
)
subparsers = parser.add_subparsers(dest="command")

# init
i = subparsers.add_parser("init")

# run
r = subparsers.add_parser("run")

# stop
s = subparsers.add_parser("stop")

# run-dev
rd = subparsers.add_parser("run-dev")

# parse args
args = parser.parse_args()


def main():
    if args.command == "init":
        cmd_pull_submodules = "git submodule update --init --recursive"
        cmd_init_api = "(cd lndb-rest/ && docker build . -t lndb-rest)"
        cmd_init_ui = "(cd lndb-ui/ && docker build . -t lndb-ui)"
        cmd = f"{cmd_pull_submodules} && {cmd_init_api} && {cmd_init_ui}"
        subprocess.run(cmd, shell=True)

    if args.command == "run":
        if not settings.instance.storage.is_cloud:
            cmd_run_api = (
                f"docker run -p 8000:8000 -v {settings_dir}:/root/.lndb -v "
                f"{settings.instance.storage.root}:{settings.instance.storage.root} "
                "--name lndb-rest lndb-rest"
            )
        else:
            cmd_run_api = (
                f"docker run -p 8000:8000 -v {settings_dir}:root/.lndb  --name"
                " lndb-rest lndb-rest"
            )
        cmd_run_ui = (
            """docker run -p 3000:3000 """
            """-e NEXT_PUBLIC_LAMIN_REST_DB_URL="http://localhost:8000" """
            """--name lndb-ui lndb-ui"""
        )
        cmd = f"{cmd_run_api} & {cmd_run_ui}"
        subprocess.run(cmd, shell=True)

    if args.command == "stop":
        cmd_stop_api = "docker stop lndb-rest && docker rm lndb-rest"
        cmd_stop_ui = "docker stop lndb-ui && docker rm lndb-ui"
        cmd = f"{cmd_stop_api} && {cmd_stop_ui}"
        subprocess.run(cmd, shell=True)

    if args.command == "run-dev":
        cmd_run_api = "(cd lndb-rest/lndb_rest && python3 main.py)"
        cmd_run_ui = "(cd lndb-ui/ && yarn dev)"
        cmd = f"{cmd_run_api} & {cmd_run_ui}"
        subprocess.run(cmd, shell=True)



import lamindb as ln
from fastapi import APIRouter
import pandas as pd
from lnschema_core.link import RunIn

router = APIRouter(prefix="/run")


@router.get("/")
async def get_runs(order_by: str = None, desc: bool = True):

    if order_by == "name":
        order_by_exp = ln.Run.name
    else:
        order_by_exp = ln.Run.created_at

    if desc is True:
        order_by_exp = order_by_exp.desc()

    runs = (
        ln.select(ln.Run.id, ln.Run, ln.User, ln.Transform)
        .join(ln.Transform, (ln.Transform.id == ln.Run.transform_id), (ln.Transform.v == ln.Run.transform_v))
        .join(ln.User, ln.User.id == ln.Run.created_by)
        .order_by(order_by_exp)
    ).all()

    if len(runs) == 0:
        return []

    input_files = (
        ln.select(ln.Run.id, ln.File)
        .join(RunIn, ln.Run.id == RunIn.run_id)
        .join(ln.File, RunIn.file_id == ln.File.id)
    ).all()

    output_files = (
        ln.select(ln.Run.id, ln.File)
        .join(ln.File, ln.File.source_id == ln.Run.id)
    ).all()

    runs_df = pd.DataFrame(runs)

    if len(input_files) > 0:
        runs_input_files_df = (
            pd.DataFrame(input_files)
            .rename(columns={"File": "Input"})
            .groupby("id")["Input"]
            .apply(list)
            .reset_index()
        )
    else:
        runs_input_files_df = pd.DataFrame([], columns=["id", "Input"])

    if len(output_files) > 0:
        runs_output_files_df = (
            pd.DataFrame(output_files)
            .rename(columns={"File": "Output"})
            .groupby("id")["Output"]
            .apply(list)
            .reset_index()
        )
    else:
        runs_output_files_df = pd.DataFrame([], columns=["id", "Output"])

    runs_df = join_with_runs_input_files(runs_df, runs_input_files_df)
    runs_df = join_with_runs_output_files(runs_df, runs_output_files_df)

    return runs_df.to_dict("records")


def join_with_runs_input_files(runs_df: pd.DataFrame, runs_input_files_df: pd.DataFrame):
    run_with_files = runs_df.merge(runs_input_files_df, how="left")
    run_with_files["Input"] = run_with_files["Input"].fillna("").apply(list)
    return run_with_files


def join_with_runs_output_files(runs_df: pd.DataFrame, runs_output_files_df: pd.DataFrame):
    run_with_files = runs_df.merge(runs_output_files_df, how="left")
    run_with_files["Output"] = run_with_files["Output"].fillna("").apply(list)
    return run_with_files


@router.get("/{run_id}")
async def get_run(run_id: str, order_by_file: str = None, desc_file: bool = True):

    if order_by_file == "name":
        order_by_file_exp = ln.File.name
    else:
        order_by_file_exp = ln.File.created_at

    if desc_file is True:
        order_by_file_exp = order_by_file_exp.desc()

    run = (
        ln.select(ln.Run, ln.User, ln.Transform)
        .join(ln.Transform, (ln.Transform.id == ln.Run.transform_id), (ln.Transform.v == ln.Run.transform_v))
        .join(ln.User, ln.User.id == ln.Run.created_by)
        .where(ln.Run.id == run_id)
    ).one()

    input_files = (
        ln.select(ln.Run.id, ln.File)
        .join(RunIn, ln.Run.id == RunIn.run_id)
        .join(ln.File, RunIn.file_id == ln.File.id)
        .order_by(order_by_file_exp)
    ).all()

    output_files = (
        ln.select(ln.Run.id, ln.File)
        .join(ln.File, ln.File.source_id == ln.Run.id)
        .order_by(order_by_file_exp)
    ).all()

    return {
        **run,
        "Input": input_files,
        "Output": output_files
    }
